#!/usr/bin/env python3

import urllib.request
import urllib.error
import json
import re

from argparse import ArgumentParser, Action
from calendar import timegm
from threading import Thread
from queue import Queue
from time import time, strftime, strptime, gmtime
from sys import exit, stderr

# Generic MirrorStatus Exception
class MirrorStatusError(Exception):
  def __init__(self, msg):
    self.msg = msg
  def __str__(self):
    return repr(self.msg)




class MirrorStatus():
  # JSON URI
  URL = 'https://www.archlinux.org/mirrors/status/json/'
  # Mirror URL format. Accepts server base URL, repository, and architecture.
  MIRROR_URL_FORMAT = '{0}{1}/os/{2}'
  MIRRORLIST_ENTRY_FORMAT = "Server = " + MIRROR_URL_FORMAT + "\n"
  DISPLAY_TIME_FORMAT = '%Y-%m-%d %H:%M:%S GMT'
  PARSE_TIME_FORMAT = '%Y-%m-%d %H:%M:%S'
  # Recognized list sort types and their descriptions.
  SORT_TYPES = {'age': 'last server synchronization',
                'rate': 'download rate',
                'country': 'server\'s location'}
  # Known repositories, i.e. those that should be on each mirror.
  # Used to replace the "$repo" variable.
  REPOSITORIES = [
    'community',
    'community-staging',
    'community-testing',
    'core',
    'extra',
    'gnome-unstable',
    'kde-unstable',
    'multilib',
    'multilib-testing'
    'staging',
    'testing']

  # Known system architectures, as used to replace the "$arch" variable.
  ARCHITECTURES = ['i686', 'x86_64']

  # Initialize
  # refresh_interval:
  #   The cached list will be replaced after this many seconds have passed.
  #   0 effectively disables caching.
  #   Caching is only useful if the object persists, e.g. if it were embedded
  #   in a server.
  def __init__(self, refresh_interval=0):
    self.refresh_interval = refresh_interval

    # Last modification time of the json object.
    self.json_mtime = 0
    # The parsed JSON object.
    self.json_obj = {}



  # Retrieve the current mirror status JSON data.
  def retrieve(self):
    try:
      req = urllib.request.Request(url=MirrorStatus.URL)
      f = urllib.request.urlopen(req)
      self.json_obj = json.loads( f.read().decode() )
      self.json_mtime = time()
      f.close()

      # Remove servers that have not synced, and parse the "last_sync" times for
      # comparison later.
      mirrors = self.json_obj['urls']
      # Filter mirrors that haven't synced.
      mirrors = list( filter(lambda m: m['last_sync'], mirrors) )
      # Parse 'last_sync' times for future comparison.
      for mirror in mirrors:
        mirror['last_sync'] = timegm(strptime(mirror['last_sync'], MirrorStatus.PARSE_TIME_FORMAT))
      self.json_obj['urls'] = mirrors

    # Missing keys mean that something went wrong, most likely a transient
    # HTTP(S) error.
    # TODO: Improve error handling (more information, handle specific cases
    #       based on return status, etc)
    except (urllib.error.URLError, ValueError, KeyError) as e:
      raise MirrorStatusError('failed to retrieve mirror data: %s' % (e))





  # Return the JSON object, retrieving new data if necessary.
  def get_obj(self):
    if not self.json_obj \
    or time() > (self.json_mtime + self.refresh_interval):
      self.retrieve()

    return self.json_obj


  # Get the mirrors
  def get_mirrors(self):
    return self.get_obj()['urls']



  # Filter
  def filter(self, mirrors=None, countries=None, regexes=None, age=None, protocols=None):
    if not mirrors:
      mirrors = self.get_mirrors()

    t = time()
    n = 0

    for mirror in mirrors:
      # Filter by country.
      if countries and not mirror['country'] in countries:
        continue
      # Filter by protocol.
      if protocols and not mirror['protocol'] in protocols:
        continue
      # Filter by regex.
      # TODO: Find a better way to do this.
      if regexes:
        for regex in regexes:
          if re.search(regex, mirror['url']):
            break
        else:
          continue
      # Filter by hours since last sync.
      if age and t > (age * 60**2 + mirror['last_sync']):
        continue

      # Yield if we're still here.
      yield mirror

  # Sort
  def sort(self, mirrors=None, by=None):
    if not mirrors:
      mirrors = self.get_mirrors()

    # Ensure that "mirrors" is a list that can be sorted.
    if not isinstance(mirrors, list):
      mirrors = list(mirrors)

    if by == 'age':
      mirrors.sort(key=lambda m: m['last_sync'], reverse=True)
    elif by == 'rate':
      mirrors = self.rate(mirrors)
    elif by == 'country':
      mirrors.sort(key=lambda m: m['country'])

    return mirrors



  # Sort mirrors by download speed. Download speed will be calculated from the
  # download time of the [core] database from each server.
  # TODO: Consider ways to improve this.
  # TODO: Consider the effects of threading (do the threads skew the results by
  #       competing for bandwidth?)
  def rate(self, mirrors=None, threads=5):
    if not mirrors:
      mirrors = self.get_mirrors()
    # Ensure that "mirrors" is a list and not a generator.
    if not isinstance(mirrors, list):
      mirrors = list(mirrors)

    # Ensure a sane number of threads.
    if threads < 1:
      threads = 1
    else:
      threads = min(threads, len(mirrors))

    rates = {}

    # URL input queue
    q_in = Queue()
    # URL and rate output queue
    q_out = Queue()
    def worker():
      while True:
        url = q_in.get()
        db_url = url + 'core/os/i686/core.db'
        req = urllib.request.Request(url=db_url)
        # Leave the rate as 0 if the connection fails.
        # TODO: Consider more graceful error handling.
        rate = 0
        try:
          t0 = time()
          core = urllib.request.urlopen(req, None, 5).read()
          t1 = time()
          rate = len(core) / (t1 - t0)
        except:
          pass
        q_out.put((url, rate))
        q_in.task_done()

    # Launch threads
    for i in range(threads):
      t = Thread(target=worker)
      t.daemon = True
      t.start()


    # Load the input queue
    for mirror in mirrors:
      q_in.put(mirror['url'])

    q_in.join()

    # Get the results
    # The "in mirrors" loop is just used to ensure that the right number of
    # items is retrieved.
    for mirror in mirrors:
      url, rate = q_out.get()
      rates[url] = rate
      q_out.task_done()

    mirrors.sort(key=lambda m: rates[m['url']], reverse=True)

    return mirrors




  # Return a Pacman-formatted mirrorlist
  # TODO: Reconsider the assumption that self.json_obj has been retrieved.
  def get_mirrorlist(self, mirrors=None, include_country=False):
    if not mirrors:
      mirrors = self.get_mirrors()

    mirrorlist = ("# Generated by Reflector\n" + "# %-10s %s\n" * 4 + "\n") % (\
      'When:',
      strftime(MirrorStatus.DISPLAY_TIME_FORMAT, gmtime()),
      'Source:',
      MirrorStatus.URL,
      'Retrieved:',
      strftime(MirrorStatus.DISPLAY_TIME_FORMAT, gmtime(self.json_mtime)),
      'Updated:',
      self.json_obj['last_check'] + ' GMT'
      )

    country = None

    for mirror in mirrors:
      # Include country tags. This is intended for lists that are sorted by
      # country.
      if include_country:
        c = mirror['country']
        if c != country:
          if country:
            mirrorlist += "\n"
          mirrorlist += "# %s\n" % c
          country = c
      mirrorlist += MirrorStatus.MIRRORLIST_ENTRY_FORMAT.format(mirror['url'], '$repo', '$arch')

    return mirrorlist




  def list_countries(self):
    countries = {}
    for m in self.get_mirrors():
      try:
        countries[m['country']] += 1
      except KeyError:
        countries[m['country']] = 1
    return countries






# List countries along with the number of mirrors in each.
class ListCountries(Action):
  def __call__(self, parser, namespace, values, option_string=None):
    ms = MirrorStatus()
    countries = ms.list_countries()
    w = 0
    for c in countries:
      w = max(w, len(c))
    for c in sorted(countries.items(), key=lambda x: x[0]):
      print (('%-'+str(w)+'s %3d') % c)
    exit(0)



# Parse arguments. If none are passed, use ARGV.
def parse_args(args=None):
  parser = ArgumentParser(description='retrieve and filter a list of the latest Arch Linux mirrors')

  parser.add_argument('--list-countries', action=ListCountries, nargs=0,
                    help='Display a table of the distribution of servers by country.')

  parser.add_argument('--save', dest='save', metavar='<filepath>',
                    help='Save the mirrorlist to the given path.')

  parser.add_argument('--sort', dest='sort', choices=MirrorStatus.SORT_TYPES,
                    help='Sort the mirrorlist: %s.' % ('; '.join(map(\
                    lambda x: "\"%s\": %s" % x, MirrorStatus.SORT_TYPES.items() \
                    ))))

  parser.add_argument('--threads', dest='threads', type=int, metavar='n',
                    help='The number of threads to use when rating mirrors.')



  filters = parser.add_argument_group('filters', 'The following filters are inclusive, i.e. the returned list will only contain mirrors for which all of the given conditions are met.')

  filters.add_argument('-a', '--age', dest='age', type=float, metavar='n',
                    help='Only return mirrors that have synchronized in the last n hours. n may be an integer or a decimal number.')

  filters.add_argument('-c', '--country', dest='countries', action='append', metavar='<country>',
                    help='Match one of the given countries (case-sensitive). Use "--list-countries" to see which are available.')

  filters.add_argument('-f', '--fastest', dest='fastest', type=int, metavar='n',
                    help='Return the n fastest mirrors that meet the other criteria.')

  filters.add_argument('--grep', dest='grep', metavar='<regex>', action='append',
                    help=' the list to URLs that match at least one of the given regular expressions.')

  filters.add_argument('-l', '--latest', dest='latest', type=int, metavar='n',
                    help='Limit the list to the n most recently synchronized servers.')

  filters.add_argument('-n', '--number', dest='number', type=int, metavar='n',
                    help='Return at most n mirrors.')

  filters.add_argument('-p', '--protocol', dest='protocols', action='append', metavar='<protocol>',
                    help='Match one of the given protocols, e.g. "http", "ftp".')



  options = parser.parse_args(args)

  return options



# Process options
def process_options(options, ms=None, mirrors=None):
  if not ms:
    ms = MirrorStatus()

  if not mirrors:
    mirrors = ms.get_mirrors()

  # Filter
  mirrors = ms.filter(mirrors,
    countries=options.countries,
    regexes=options.grep,
    age=options.age,
    protocols=options.protocols)

  if options.latest and options.latest > 0:
    mirrors = ms.sort(mirrors, 'age')
    mirrors = mirrors[:options.latest]

  if options.fastest and options.fastest > 0:
    mirrors = ms.sort(mirrors, 'rate')
    mirrors = mirrors[:options.fastest]

  if options.sort:
    mirrors = ms.sort(mirrors, options.sort)

  if options.number:
    mirrors = list(mirrors)[:options.number]

  return ms, mirrors





def main():
  options = parse_args()

  try:
    ms, mirrors = process_options(options)
    include_country = options.sort == 'country'
    mirrorlist = ms.get_mirrorlist(mirrors, include_country=include_country)
  except MirrorStatusError as e:
    stderr.write('error: %s' % e.msg)
    exit(1)

  if options.save:
    try:
      f = open(options.save, 'w')
      f.write(mirrorlist)
      f.close()
    except IOError as e:
      stderr.write(e)
      exit(1)
  else:
    print(mirrorlist)



if __name__ == "__main__":
  main()


